"""
Based on the :py:mod:`pylint.pyreverse.main` module.
Under Debian, see the
``/usr/lib/python3/dist-packages/pylint/pyreverse/main.py``

**Background:**

We must overlad the :py:class:`pylint.pyreverse.main.Run` class as we need
an extra option to set the directory where Sphinx stores its output HTML files.

However, adding this option is not straightforward.
The :py:meth:`pylint.pyreverse.main.Run.__init__` constructor is monolotic
and runs the following operations:

- parsing the arguments from the CLI;
- initializing the configuration needed by the :py:meth:`pyreverse.run` method;
- calling the :py:meth:`pylint.pyreverse.main.Run.run` method with the
  remaining arguments;
- exiting the program using the :py:func:`sys.exit` function!

This design prevents to call `pyreverse` multiple times, which is needed to
process by a Sphinx extension producing muliple UML diagrams.

To address this problem, this file scatters those steps as follows:

- The :py:class:`ParsePyreverseArgs` class parses the arguments;
- The :py:class:`Run.__init__` constructor takes in parameter the configuration
  obtained from the :py:class:`ParsePyreverseArgs` class;
- The :py:class:`Run.run` constructor takes in parameter the remaining
  CLI arguments, obtained from the :py:attr:`ParsePyreverseArgs.remaining_args`
  attribute;
"""

# Inherited imports
from pylint.pyreverse.main import (
    augmented_sys_path,
    check_graphviz_availability,
    check_if_graphviz_supports_format,
    DiadefsHandler,
    DIRECTLY_SUPPORTED_FORMATS,
    # discover_package_path,
    insert_default_options,
    Linker,
    project_from_files,
    writer,
)

from pylint.config.arguments_manager import _ArgumentsManager
from pylint.config.arguments_provider import _ArgumentsProvider

# Custom imports
import argparse
from importlib.metadata import version
from packaging.version import Version
from pylint.pyreverse.main import Run as _Run, OPTIONS as _OPTIONS
from .dot_printer import DotPrinter
from .sphinx_html_proxy import SphinxHtmlProxy


# << Added options
OPTIONS = [
    option
    for option in sorted(list(_OPTIONS) + [
        (
            "sphinx-html-dir",
            {
                "default": "",
                "type": "path",
                "short": "d",
                "action": "store",
                "metavar": "<input_directory>",
                "help": (
                    "set the root directory of the HTML "
                    "documentation generated by Sphinx."
                ),
            },
        )
    ])
]
# >>


class ParsePyreverseArgs(_Run):
    options = OPTIONS
    name = "pyreverse"

    # << Based on _Run.__init__
    def __init__(self, args: list[str]):
        """
        Constructor.
        Based on :py:meth:`pylint.pyreverse.main.Run.__init__`.

        Args:
            args (list[str]): The arguments passed to the script.
                *Example:* ``sys.argv[1:]``.
        """
        _ArgumentsManager.__init__(self, prog="pyreverse", description=__doc__)
        _ArgumentsProvider.__init__(self, self)

        # Parse options
        insert_default_options()
        # << We need to save them to call Run.run(...)
        self.remaining_args = self._parse_command_line_configuration(args)
        # >>

        if self.config.output_format not in DIRECTLY_SUPPORTED_FORMATS:
            check_graphviz_availability()
            print(
                f"Format {self.config.output_format} isn't supported natively."
                " pyreverse2 will try to generate it using Graphviz..."
            )
            check_if_graphviz_supports_format(self.config.output_format)


class Run:
    def __init__(self, config: argparse.Namespace):
        """
        Constructor.

        Args:
            config (argparse.Namespace): The configuration parsed from the
                command line. See also the ;py:class;`ParsePyreverseArgs`
                class.

        Returns:
            The execution code (``0`` means everything is fine).
        """
        # pylint.pyreverse.main.Run is meant to be called only once
        # so let's rewrite __init__:
        self.config = config

    def diadefs(self, args: list[str]):
        # Squeeze extra_packages_paths, because it explores src/
        # TODO: We should see what pyreverse does
        # (especially, what is the value of self.config.source_roots)
        # <<
        # extra_packages_paths = list({
        #     discover_package_path(
        #         arg,                      # example.module.submodule.c1
        #         self.config.source_roots  # ()
        #     )
        #     for arg in args
        # })
        # ===
        extra_packages_paths = list()
        # >>

        with augmented_sys_path(extra_packages_paths):
            project = project_from_files(
                args,
                project_name=self.config.project,
                black_list=self.config.ignore_list,
                verbose=self.config.verbose,
            )
            linker = Linker(project, tag=True)
            if Version(version("pylint")) < Version("4.0.0"):
                handler = DiadefsHandler(self.config)
            else:
                handler = DiadefsHandler(self.config, args=list())
            diadefs = handler.get_diadefs(project, linker)
            return diadefs

    def run(self, args: list[str]) -> int:
        """
        Checks the arguments and pyreverses the input project.

        Args:
            args (list[str]): The remaining arguments, that are not yet
                handled by the constructor, typically, the class or module
                being pyreversed.
                *Example:* ``['example.module.submodule.c1']``.
        """
        dwriter = writer.DiagramWriter(self.config)
        # << Writer hijacking
        dwriter.printer_class = DotPrinter
        dwriter.api_doc = SphinxHtmlProxy()
        dwriter.api_doc.set_sphinx_html_dir(self.config.sphinx_html_dir)
        # NB: dwriter.write calls DotPrinter.__init__(...)
        diadefs = self.diadefs(args)
        # >>
        dwriter.write(diadefs)
        return 0



from .cli import run_pyreverse2  # noqa: F401
from .dot_printer import DotPrinter  # noqa: F401
from .main import ParsePyreverseArgs, Run  # noqa: F401
from .sphinx_html_proxy import SphinxHtmlProxy  # noqa: F401



"""
Based on the :py:mod:`pyreverse.dot_printer` module.
Under Debian, see the
``/usr/lib/python3/dist-packages/pylint/pyreverse/dot_printer.py``
file.
"""

# Inherited imports
from pylint.pyreverse.dot_printer import (
    NodeProperties,
    nodes,
    get_annotation_label,
    NodeType,
    SHAPES
)

# Custom imports
from pylint.pyreverse.dot_printer import DotPrinter as _DotPrinter
from .sphinx_html_proxy import SphinxHtmlProxy


class DotPrinter(_DotPrinter):
    """
    Overloads the :py:class:`pylint.pyreverse.dot_printer.DotPrinter` class
    to export an UML diagram using Graphviz and crafting links in the output
    file to map each class/method/attribute name with the corresponding Sphinx
    HTML page/anchor.
    """
    def _open_graph(self) -> None:
        super()._open_graph()
        self.emit("bgcolor=transparent")

    def _build_label_for_node(  # noqa: C901 func too-complex
        self,
        properties: NodeProperties
    ) -> str:
        if not properties.label:
            return ""

        # Attribute and method display
        # HTML tags supported by Graphviz:
        # https://graphviz.org/doc/info/shapes.html#html
        # NB: A node can embed at most one <table> tag.
        # Several links in a single node: https://stackoverflow.com/a/48029398

        def vstack(**kwargs: dict) -> str:
            text = kwargs.pop("text", None)
            if not text:
                return ""
            if "tooltip" not in kwargs:
                kwargs["tooltip"] = ""
            if "href" in kwargs and kwargs["href"] is None:
                kwargs.pop("href")
            attrs = " ".join([
                f'{k}="{v}"'
                for k, v in kwargs.items()
            ])
            return f"\n\t<tr><td {attrs}>{text}</td></tr>"

        def escape_string(s: str) -> str:
            return s.replace('"', '\\"')

        proxy = SphinxHtmlProxy()
        class_name = properties.label
        label = (
            f'<table border="0" align="left" tooltip="{class_name}" '
            'width="0" cellpadding="0">'
        )
        class_url = proxy.url()

        # Class name
        label += vstack(
            border=1,
            href=class_url,
            tooltip=escape_string(class_name),
            target="_top",
            text=f"<b>{class_name}</b>"
        )

        # Only class names
        if properties.attrs is None and properties.methods is None:
            return label + "</table>"

        attrs: list[str] = properties.attrs or []
        methods: list[nodes.FunctionDef] = properties.methods or []

        # Add class attributes
        if attrs:
            label += vstack(
                align="left",
                text="<b>Attributes:</b>"
            )
        for attr in attrs:
            attr_url = proxy.url(attr)
            attr_label = attr.replace("|", r"\|")
            label += vstack(
                align="left",
                href=attr_url,
                target="_top",
                tooltip=escape_string(f"{class_name}.{attr_label}"),
                text=attr_label
            )

        # Add class methods
        if methods:
            label += vstack(
                align="left",
                text="<b>Methods:</b>"
            )
        for func in methods:
            args = (
                ", "
                .join(self._get_method_arguments(func))
                .replace("|", r"\|")
            )
            method_name = func.name
            method_url = proxy.url(func.name)
            prototype = rf"{method_name}({args})"
            if func.returns:
                annotation_label = get_annotation_label(func.returns)
                prototype += (
                    ": " + self._escape_annotation_label(annotation_label)
                )
            if func.is_abstract():
                prototype = f"<i>{prototype}</i>"
            label += vstack(
                align="left",
                href=method_url,
                target="_top",
                tooltip=escape_string(f"{class_name}.{func.name}"),
                text=prototype
            )

        label += "</table>"
        # >>
        return label

    def emit_node(
        self,
        name: str,
        type_: NodeType,
        properties: NodeProperties | None = None,
    ) -> None:
        proxy = SphinxHtmlProxy()
        i = name.rfind(".")
        if i > 0:
            proxy.module_name = name[: i]
            proxy.class_name = name[i + 1:]
        else:
            proxy.module_name = name
            proxy.class_name = None

        if properties is None:
            properties = NodeProperties(label=name)
        shape = SHAPES[type_]
        color = (
            properties.color if properties.color is not None
            else self.DEFAULT_COLOR
        )
        style = "filled" if color != self.DEFAULT_COLOR else "solid"
        label = self._build_label_for_node(properties)
        label_part = f", label=<{label}>" if label else ""
        fontcolor_part = (
            f', fontcolor="{properties.fontcolor}"' if properties.fontcolor
            else ""
        )

        # URLs in Graphviz: https://graphviz.org/docs/attrs/URL/
        # https://talk.observablehq.com/t/hyperlink-in-graphviz-node-doesnt-work/4775/2
        # Example:
        # dot`digraph { b [URL="https://google.com" target="_top"]; a -> b; }`
        url = proxy.url()
        self.emit(
            f'"{name}" [' + ", ".join([
                f'color="{color}"{fontcolor_part}{label_part}',
                f'shape="{shape}"',
                f'style="{style}"',
                f'URL="{url}"',
            ]) + "];"
        )



__author__ = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__ = "marc-olivier.buob@nokia-bell-labs.com"
__license__ = "BSD-3"
__version__ = "0.4.0"
__all__ = ["UMLGenerateDirective"]


import os
from sphinx.application import Sphinx
from sphinx.util.typing import ExtensionMetadata
from .uml_generate_directive import UMLGenerateDirective


def setup(app: Sphinx) -> ExtensionMetadata:
    """Setup directive"""
    app.setup_extension("sphinx.ext.graphviz")

    # Allow override of the directive, defaulting to "uml"
    directive_name_to_use = os.environ.get("SPHINX_UML_DIRECTIVE", "uml")
    app.add_directive(directive_name_to_use, UMLGenerateDirective)

    # sphinx.ext.inheritance_diagram-like options
    # app.add_config_value("uml_graph_attrs", {}, "")
    # app.add_config_value("uml_node_attrs", {}, "")
    # app.add_config_value("uml_edge_attrs", {}, "")
    # app.add_config_value("uml_alias", {}, "")

    # pylint.pyreverse-like options
    app.add_config_value("uml_filter_mode", default=None, rebuild="env")
    app.add_config_value("uml_class", default=None, rebuild="env")
    app.add_config_value("uml_show_ancestors", default=None, rebuild="env")
    app.add_config_value("uml_all_ancestors", default=None, rebuild="env")
    app.add_config_value("uml_show_associated", default=None, rebuild="env")
    app.add_config_value("uml_all_associated", default=None, rebuild="env")
    app.add_config_value("uml_show_builtin", default=None, rebuild="env")
    app.add_config_value("uml_module_names", default=None, rebuild="env")
    app.add_config_value("uml_only_classnames", default=None, rebuild="env")
    app.add_config_value("uml_ignore", default=None, rebuild="env")
    app.add_config_value("uml_colorized", default=None, rebuild="env")

    return {
        "version": __version__,
        "parallel_read_safe": True
    }



from pathlib import Path
from ..singleton import Singleton


class SphinxHtmlProxy(metaclass=Singleton):
    """
    The :py:class:`SphinxHtmlProxy` is a singleton in charge of returning
    the path of an HTML file corresponding to a class of the project being
    pyreversed.
    """
    def __init__(self):
        """
        Constructor. The attributes are set by the calling code,
        depending on the context.
        """
        self.sphinx_html_dir: str  # The HTML Spring address or directory
        self.module_name: str      # The module being processed
        self.class_name: str       # The class being processed

    def set_sphinx_html_dir(self, sphinx_html_dir: Path | str):
        """
        Set the HTML Sphinx documentation directory.

        Args:
            sphinx_html_dir (Path | str): The input directory.
                *Example:* ``~/git/unext/docs/_build/html``.
        """
        # assert Path(sphinx_html_dir).exists(), sphinx_html_dir
        self.sphinx_html_dir = sphinx_html_dir

    def url(self, name: str = "") -> str:
        """
        Crafts an Sphinx URL depending on:

        - the HTML Sphinx directory (:py:attr:`sphinx_html_dir`);
        - the current module (:py:attr:`module_name`);
        - the current class (:py:attr:`class_name`);
        - the current method/attribute if any (see ``name``).

        Args:
            name (str): The current method/attribute name if any,
                ``""`` otherwise.

        Returns:
            The resulting URL if any, ``None`` otherwise.
        """
        if not self.sphinx_html_dir:
            # HTML documentation directory not set.
            return None
        if ":" in name:
            name = name[:name.find(":")]
        return (
            # File
            f"{self.sphinx_html_dir}/{self.module_name}.html"
            # Anchor (class)
            f"#{self.module_name}.{self.class_name}"
        ) + (
            # Anchor (attr / method)
            f".{name}" if name
            else ""
        )



#!/usr/bin/env python3

"""
Based on the :py:mod:`pylint.pyreverse.main` module.
Under Debian, see the
``/usr/lib/python3/dist-packages/pylint/__init__.py``
file.
"""

import sys
from .main import ParsePyreverseArgs, Run


def run_pyreverse2(args: list[str] = None):
    """
    ``pyreverse2`` entry point, which is used to draw
    UML diagrams by inspecting some python module.

    To display the help related to this program:

    .. code:: bash

        pyreverse2 --help

    To run this program:

    .. code:: bash

        pyreverse2 \\
            --output svg \\
            --project example.a \\
            --sphinx-html-dir docs/_html \\
            --output-directory docs/ \\
            -m y \\
            example.a

    Args:
        args (list[str]): The arguments passed to the script.
            *Example:* ``sys.argv[1]``.
    """
    args = args or sys.argv[1:]
    parser = ParsePyreverseArgs(args)
    runner = Run(parser.config)
    runner.run(parser.remaining_args)


if __name__ == "__main__":
    run_pyreverse2()



import argparse
from pathlib import Path
from docutils import nodes
from docutils.parsers.rst import directives
from pylint.pyreverse.main import writer
from pylint.pyreverse.diagrams import (
    ClassDiagram,
    PackageDiagram,
)
from sphinx.ext.graphviz import (
    figure_wrapper,
    graphviz,
)
from sphinx.util.docutils import SphinxDirective
from sphinx.util.typing import OptionSpec
from typing import ClassVar


class UmlNode(graphviz):
    """
    Defines a UML ``docutils`` node.
    We populate its attribute so that we can rely on the
    default ``graphviz`` node.
    """
    @classmethod
    def from_dot(cls, dotcode: str) -> "UmlNode":
        """
        Builds a :py:class:`UmlNode` instance from a Graphviz
        dot string.

        Args:
            dotcode (str): A Graphviz dot string.
                *Example:* ``digraph G {0 -> 1}``

        Returns:
            The resulting :py:class:`UmlNode` instance.
        """
        node = cls()
        node["code"] = dotcode
        node["options"] = {"graphviz_dot": "dot"}

        # We rely on sphinx.ext.inheritance_diagram CSS classes
        # to be responsive to dark/light themes.
        # TODO: improve this!
        node["classes"] = ["inheritance"]
        return node

    @classmethod
    def to_dot(
        cls,
        diagram: ClassDiagram | PackageDiagram,
        config: argparse.Namespace
    ) -> str:
        """
        Exports a diagram definition obtained from ``pyreverse``
        to a Graphviz dot string.

        Args:
            diagram (ClassDiagram | PackageDiagram): The diagram
                that must be exported.
            config (argparse.Namespace): The configuration
                obtained from the Sphinx configuration file.

        Returns:
            The resulting :py:class:`UmlNode` instance.
        """
        from .pyreverse import DotPrinter, SphinxHtmlProxy
        dwriter = writer.DiagramWriter(config)
        dwriter.printer_class = DotPrinter
        # TODO Build xrefs as in
        # /usr/lib/python3/dist-packages/sphinx/ext/inheritance_diagram.py?
        dwriter.api_doc = SphinxHtmlProxy()
        dwriter.api_doc.sphinx_html_dir = config.sphinx_html_dir
        dwriter.write([diagram])
        return "\n".join(dwriter.printer.lines)

    @classmethod
    def from_pyreverse(
        cls,
        diagram: ClassDiagram | PackageDiagram,
        config: argparse.Namespace
    ) -> "UmlNode":
        """
        Builds a :py:class:`UmlNode` from a diagram definition
        obtained from ``pyreverse``.

        Args:
            diagram (ClassDiagram | PackageDiagram): The diagram
                that must be exported.
            config (argparse.Namespace): The configuration
                obtained from the Sphinx configuration file.

        Returns:
            The resulting :py:class:`UmlNode` instance.
        """
        return cls.from_dot(cls.to_dot(diagram, config))


def guess_svg_basename(options: dict, code: str) -> str:
    """
    Infers the svg filename of an UML diagram
    See ``/usr/lib/python3/dist-packages/sphinx/ext/graphviz.py``.

    Args:
        options (dict): The options passed to the
            :py:func:`graphviz.ext.render_dot` function.
        code (str): A graphviz string, passed to the
            :py:func:`graphviz.ext.render_dot` function.

    Returns:
        The corresponding SVG basename
    """
    from hashlib import sha1
    graphviz_dot = options.get("graphviz_dot", None)
    hashkey = (code + str(options) + str(graphviz_dot) + "()").encode()
    svg_basename = (
        f'graphviz-{sha1(hashkey, usedforsecurity=False).hexdigest()}.svg'
    )
    return svg_basename


class UMLGenerateDirective(SphinxDirective):
    """
    UML directive to generate a pyreverse diagram
    """

    # Sphinx stuff to control argument passing
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec: ClassVar[OptionSpec] = {
        "caption": directives.unchanged,
        "classes": directives.flag,
        "packages": directives.flag,
    }

    def _build_args(self) -> list[str]:
        args = list()
        config = self.config
        if config.uml_filter_mode:
            assert config.uml_filter_mode
            args.extend(("--filter-mode", config.uml_filter_mode))
        if config.uml_class:
            args.extend(("--class", config.uml_class))
        if config.uml_show_ancestors:
            args.extend(("--show-ancestors", config.uml_show_ancestors))
        if config.uml_all_ancestors:
            args.append("--all-ancestors")
        if config.uml_show_associated:
            args.extend(("--show-associated", config.uml_show_associated))
        if config.uml_all_associated:
            args.append("--all-associated")
        if config.uml_show_builtin:
            args.append("--show-builtin")
        if config.uml_module_names:
            args.extend(("--module-names", config.uml_module_names))
        if config.uml_only_classnames:
            args.append("--only-classnames")
        if config.uml_ignore:
            args.extend(("--ignore", config.uml_ignore))
        if config.uml_colorized:
            args.append("--colorized")
        return args

    def html_root_dir(self) -> Path:
        """
        Crafts the HTML prefix to move from the current HTML
        to the HTML root directory.

        Returns:
            The corresponding relative :py:class:`Path` instance.

        Example:
            Assume that:

            - the documentation is built in:
              ``"~/git/sphinx-uml/docs"``;
            - the current document is;
              ``"~/git/sphinx-uml/docs/users/examples.rst"``;

            Then, the returned value is ``"../"``.
            As the HTML hierarchy follows the RST hierarchy, we use
            this prefix to setup our :py:class:`SphinxHtmlProxy`.
        """
        doc = self.state.document
        env = doc.settings.env
        base_dir = Path(env.srcdir).absolute()
        cur_path = Path(doc.current_source)
        rel_path = str(cur_path.relative_to(base_dir).parent)
        if rel_path == ".":
            return Path(rel_path)
        n = len(rel_path.split("/"))
        return Path("/".join([".."] * n))

    def run(self):
        """
        To test this extension, as a developer:

        .. shell

            make install-sphinx-custom clean-doc docs

        To test this `pyreverse2`, called by this extension:

        .. shell

            pyreverse2 \\
               --output svg \\
               --project example.a \\
               --sphinx-html-dir docs/_html \\
               --output-directory docs/ \\
               -m y \\
               example.a
        """
        # ..uml: module_name
        module_name = self.arguments[0]

        # :classes:, :packages:, :caption:
        with_classes = "classes" in self.options
        with_packages = "packages" in self.options
        caption = self.options.get("caption")

        pyprocess_args = self._build_args() + [
            "--sphinx-html-dir", str(self.html_root_dir()),
            module_name
        ]

        # make install-sphinx-custom clean-doc docs
        from .pyreverse import Run, ParsePyreverseArgs

        parser = ParsePyreverseArgs(pyprocess_args)
        runner = Run(parser.config)
        diadefs = runner.diadefs(parser.remaining_args)

        # Craft the list of nodes to be appended to the doctree's AST.
        ret = list()
        for diagram in diadefs:
            # :classes: and :packages: switches
            if isinstance(diagram, PackageDiagram):
                if not with_packages:
                    continue
            elif isinstance(diagram, ClassDiagram):
                if not with_classes:
                    continue
            else:
                raise ValueError(f"Invalid type {type(diagram)}")

            # Build graphviz node
            code = UmlNode.to_dot(diagram, runner.config)
            node = UmlNode.from_dot(code)

            svg_basename = guess_svg_basename(node.attributes["options"], code)

            # Appends a link 'Open in a new tab'
            paragraph = nodes.paragraph(text="")
            paragraph += nodes.reference(
                # See the HTMLTranslator.visit_reference function in
                # /usr/lib/python3/dist-packages/docutils/writers/_html_base.py
                text="Open in a new tab",
                refuri=str(
                    self.html_root_dir() / "_images" / svg_basename
                ),
            )

            # Add caption
            if caption:
                node = figure_wrapper(self, node, caption)

            ret += [node, paragraph]

        return ret



class Singleton(type):
    """
    The :py:class:`Singleton` allows to define singleton classes, i.e.
    classes that can be instantiated at most once.

    >>> class MyClass(metaclass=Singleton): pass
    >>> x = MyClass()
    >>> y = MyClass()
    >>> x is y
    True

    Based on this `thread
    <https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python>`__.
    """
    s_instances = dict()

    def __call__(cls, *args, **kwargs) -> type:
        if cls not in cls.s_instances:
            cls.s_instances[cls] = (
                super(Singleton, cls).__call__(*args, **kwargs)
            )
        return cls.s_instances[cls]
